#!BPY

#
# HL2 SMD Blender Scripts
#
# Copyright (C) 2005, Plss Roland ( roland@rptd.dnsalias.net )
# 
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation; either 
# version 2 of the License, or (at your option) any later 
# version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
Name: 'Half-Life 2 (.smd)...'
Blender: 236
Group: 'Export'
Submenu: 'Static Mesh...' export_static
Submenu: 'Animation Sequence...' export_anim
Tooltip: 'Export selected mesh to Half-Life 2 File Format (.smd)'
"""

__author__ = 'Plss Roland'
__url__ = ('Project Epsylon, http://rptd.dnsalias.net/epsylon/', 'My Homepage, http://rptd.dnsalias.net/')
__version__ = '1.0'
__bpydoc__ = """This script exports meshes into the Half-Life 2 ASCII format SMD.
It can create different kinds of SMD files depending on your needs.

Static Mesh
-----------
Exports the selected mesh as a static mesh usable for props.
If the selected mesh is parented to an armature it is used
to export skeleton data as well. Otherwise no skeleton is
exported.

Animation Sequence
------------------
Exports one animation sequences derived from data found in
an action specified by the user. The action is sampled at
fixed time intervalls and exported. You can export one
action at the time only.



Bugs
----
Bugs? Those are random features ;=)

Credits
-------
Major part of the code, especially the framework has been
taken from the 'Drag[en]gine' Game Engine project
( http://rptd.dnsalias.net/epsylon/dragengine.php )
"""

import Blender
from Blender.Mathutils import *
import struct, os, cStringIO, time, re
import math
from math import *


# constants
PI = 3.14159265
HALF_PI	= PI / 2.0
ONE_PI = PI / 180.0

# transformation matrix from blender to dragengine
transformBlenderToDE = Matrix( [-1,0,0,0], [0,0,-1,0], [0,1,0,0], [0,0,0,0] )
# transformation matrix from dragengine to hl2
transformDEToHL2 = Matrix( [-1,0,0,0], [0,0,1,0], [0,-1,0,0], [0,0,0,0] )
# transformation matrix from blender to hl2
transformBlenderToHL2 = transformBlenderToDE * transformDEToHL2



# class definitions ( based on Drag[en]gine )
#############################################
class DEBone:
	def __init__(self, name, index):
		self.name = name
		self.index = index
		self.pos = Vector( [0,0,0] )
		self.rot = Vector( [0,0,0] )
		self.restMat = Matrix( [1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1] )
		self.parent = -1
	def __repr__(self):
		return '[' + str(self.index) + ': name=' + self.name + ', parent=' + str(self.parent) + ', pos(' + str(self.pos) + '), rot(' + str(self.rot) + ') ]'
	def __str__(self):
		return self.__repr__()

class DEWeight:
	def __init__(self, index, weight):
		self.index = index
		self.weight = weight
	def __repr__(self):
		return '[ bone=' + str(self.index) + ', weight=' + str(self.weight) + ' ]'
	def __str__(self):
		return self.__repr__()
	
class DEVertex:
	def __init__(self, index):
		self.edges = []
		self.faces = []
		self.position = Vector( [0,0,0] )
		self.weights = []
		self.index = index
	def __repr__(self):
		return '[' + str(self.index) + ': edges(' + str(self.edges) + '), faces(' + str(self.faces) + ') ]'
	def __str__(self):
		return self.__repr__()
	def findEdge(self, edges, v2):
		for edge in self.edges:
			if edges[edge].vertices[1] == v2:
				return edge
		return -1
	def getHardEdgeCount(self, edges):
		count = 0
		for edge in self.edges:
			if edges[edge].hard:
				count = count + 1
		return count

class DEEdge:
	def __init__(self, v1, v2, index):
		self.vertices = [ v1, v2 ]
		self.faces = [ -1, -1 ]
		self.index = index
		self.hard = False
	def __repr__(self):
		return '[' + str(self.index) + ': vertices(' + str(self.vertices[0]) + ', ' + str(self.vertices[1]) + '), hard=' + str(self.hard) + ', faces(' + str(self.faces) + ') ]'
	def __str__(self):
		return self.__repr__()

class DEFace:
	def __init__(self, verts, index):
		self.vertices = [ verts[0].index, verts[1].index, verts[2].index ]
		self.normals = [ -1, -1, -1 ]
		self.edges = [ -1, -1, -1 ]
		self.index = index
		if len( verts ) == 4:
			self.vertices.append( verts[3].index )
			self.normals.append( -1 )
			self.edges.append( -1 )
	def __repr__(self):
		return '[' + str(self.index) + ': vertices(' + str(self.vertices) + '), normals(' + str(self.normals) + ', edges(' + str(self.edges) + ') ]'
	def __str__(self):
		return self.__repr__()
	def findCorner(self, vertex):
		for index in range( len(self.vertices) ):
			if self.vertices[index] == vertex:
				return index
		return -1
	def setNormalFor(self, vertex, normal):
		for index in range( len(self.vertices) ):
			if self.vertices[index] == vertex:
				self.normals[index] = normal



# base exporter class (from Drag[en]gine)
#########################################
class DEBaseExporter:
	
	# constructor
	def __init__(self):
		self.objMesh = None
		self.objArmature = None
		self.mesh = None
		self.armature = None
		self.refAction = None
		self.bones = []
		self.vertices = []
		self.edges = []
		self.faces = []
		self.triCount = 0
		self.quadCount = 0
		self.edgeCount = 0
		self.normalCount = 0
		self.multiFoldMesh = False
		self.degeneratedFaceCount = 0
		self.scene = None
		self.context = None
		self.oldFrame = 0
		self.oldAction = None
		self.file = None
	
	# find mesh, armature and ref action
	def initFindMeshArmRef(self):
		objects = Blender.Object.GetSelected()
		# try to find an armature for bones information and a messh
		for obj in objects:
			if obj.getType() == 'Mesh':
				if not self.objMesh:
					self.objMesh = obj
					self.mesh = Blender.NMesh.GetRaw( self.objMesh.data.name )
					if obj.getParent() and obj.getParent().getType() == 'Armature':
						if not self.objArmature:
							self.objArmature = obj.getParent()
							self.armature = Blender.Armature.Get( self.objArmature.data.name )
			elif obj.getType() == 'Armature':
				if not self.objArmature:
					self.objArmature = obj
					self.armature = Blender.Armature.Get( self.objArmature.data.name )
		# find reference action if we have an armature
		reRefAction = re.compile('reference(.ignore)?$')
		actions = Blender.Armature.NLA.GetActions()
		for action in actions.values():
			if reRefAction.match(action.getName()):
				self.refAction = action
	
	# process face corner
	def initProcessFaceCorner(self, vi, vo, c1, c2, fi):
		vo[c1].faces.append( fi )
		edge = vo[c1].findEdge( self.edges, vi[c2] )
		if edge == -1:
			edge = vo[c2].findEdge( self.edges, vi[c1] )
		theEdge = self.edges[edge]
		if theEdge.faces[0] == -1:
			theEdge.faces[0] = fi
		elif theEdge.faces[1] == -1:
			theEdge.faces[1] = fi
		else:
			self.multiFoldMesh = True
		self.faces[fi].edges[c1] = edge
	
	# add bones, vertices, faces and edges
	def initAddVertFaceEdge(self):
		# add bones
		if self.armature:
			for index in range( len(self.armature.bones) ):
				self.bones.append( DEBone( self.armature.bones[index].name, index ) )
		# add vertices
		if self.mesh:
			for index in range( len(self.mesh.verts) ):
				newVert = DEVertex(index)
				influences = self.mesh.getVertexInfluences( index )
				if not influences:
					influences = [ (self.bones[0].name, 1.0) ]
				for infl in influences:
					if infl[1] <= 0.01:
						continue
					for bi in range( len(self.bones) ):
						if self.bones[bi].name == infl[0]:
							newVert.weights.append( DEWeight( bi, infl[1] ) )
							break
				self.vertices.append( newVert )
		# add edges
		if self.mesh:
			if not self.mesh.edges:
				self.mesh.addEdgesData()
			for index in range( len(self.mesh.edges) ):
				edge = self.mesh.edges[index]
				newEdge = DEEdge( edge.v1.index, edge.v2.index, index )
				if edge.crease > 128:
					newEdge.hard = True
				self.edges.append( newEdge )
				self.vertices[ edge.v1.index ].edges.append( index )
				self.vertices[ edge.v2.index ].edges.append( index )
		# add faces
		if self.mesh:
			for index in range( len(self.mesh.faces) ):
				face = self.mesh.faces[index]
				newFace = DEFace(face.v, index)
				self.faces.append( newFace )
				# gather vertices and indices
				vi = [ face.v[0].index, face.v[1].index, face.v[2].index ]
				vo = [ self.vertices[ vi[0] ], self.vertices[ vi[1] ], self.vertices[ vi[2] ] ]
				if len( face.v ) == 4:
					vi.append( face.v[3].index )
					vo.append( self.vertices[ vi[3] ] )
				# corners
				count = len( face.v )
				for c in range(count):
					self.initProcessFaceCorner(vi, vo, c, (c + 1) % count, index)
	
	# calculate corner normals
	def initCalcCornerNormals(self):
		self.normalCount = 0
		for vert in self.vertices:
			hardEdgeCount = vert.getHardEdgeCount( self.edges )
			if hardEdgeCount == 0:
				for face in vert.faces:
					self.faces[face].setNormalFor( vert.index, self.normalCount )
				self.normalCount = self.normalCount + 1
			else:
				changed = True
				while changed:
					changed = False
					for index in vert.faces:
						face = self.faces[index]
						c = face.findCorner( vert.index )
						if face.normals[c] == -1:
							edge = self.edges[ face.edges[c] ]
							if edge.faces[0] == face.index:
								neighbor = edge.faces[1]
							else:
								neighbor = edge.faces[0]
							if edge.hard or neighbor == -1:
								face.normals[c] = self.normalCount
								self.normalCount = self.normalCount + 1
								changed = True
							else:
								neighborFace = self.faces[ neighbor ]
								neighborCorner = neighborFace.findCorner( vert.index ) 
								if neighborFace.normals[ neighborCorner ] != -1:
									face.normals[c] = neighborFace.normals[ neighborCorner ]
									changed = True
	
	# calculate informational numbers
	def initCalcInfoNumbers(self):
		if self.mesh:
			for face in self.mesh.faces:
				if len(face.v) == 3:
					self.triCount = self.triCount + 1
				elif len(face.v) == 4:
					self.quadCount = self.quadCount + 1
				else:
					self.degeneratedFaceCount = self.degeneratedFaceCount + 1
			if self.mesh.edges:
				self.edgeCount = len( self.mesh.edges )
	
	# prints out some debugging informations about the mesh
	def printMeshDebug(self):
		print 'bones:'
		for bone in self.bones:
			print bone
		print 'vertices:'
		for index in range( len(self.vertices) ):
			print self.vertices[index], '=>', self.mesh.verts[index].co
		print 'edges:'
		for edge in self.edges:
			print edge
		print 'faces:'
		for face in self.faces:
			print face
	
	# find objects to export
	def initExporterObjects(self):
		self.initFindMeshArmRef()
		self.initAddVertFaceEdge()
		self.initCalcCornerNormals()
		#self.printMeshDebug()
		self.initCalcInfoNumbers()
	
	# store old values to revert to them after we finished or something failed
	def storeOldValues(self):
		self.scene = Blender.Scene.getCurrent()
		self.context = self.scene.getRenderingContext()
		self.oldFrame = self.context.currentFrame()
		if self.armature:
			self.oldAction = self.objArmature.getAction()
	
	# restore old values
	def restoreOldValues(self):
		self.scene.getRenderingContext().currentFrame(self.oldFrame)
		if self.objArmature and self.oldAction:
			self.oldAction.setActive(self.objArmature)
		self.scene.update(True)
	
	# switch to given action and time
	def switchActionTime(self, action, time):
		if self.objArmature:
			action.setActive(self.objArmature)
		self.context.currentFrame(time)
		self.scene.update(True)
	
	# prints out to the console some informations
	def printInfos(self):
		if self.objMesh:
			print '*** Mesh ***'
			print 'name:    ', self.objMesh.data.name
			print 'bones:   ', len( self.bones )
			print 'textures:', len( self.mesh.materials )
			print 'vertices:', len( self.mesh.verts )
			print 'faces:   ', (self.triCount + self.quadCount)
			print '- tris:  ', self.triCount
			print '- quads: ', self.quadCount
			print 'edges:   ', self.edgeCount
			print 'normals: ', self.normalCount
		if self.objArmature:
			print '*** Armature ***'
			print 'name:    ', self.objArmature.data.name
			print 'bones:   ', len( self.armature.bones )
	
	# export file
	def export(self, filename):
		self.initExporterObjects()
		if not self.checkInitState():
			return False
		self.file = open(filename, "w")
		self.storeOldValues()
		try:
			result = self.safeExport()
		finally:
			self.restoreOldValues()
			self.file.close()
		self.printInfos()
		return result



# static mesh exporter
########################
class HLExporterStaticMesh ( DEBaseExporter ):
	
	# constructor
	def __init__(self):
		DEBaseExporter.__init__(self)
	
	# title of export dialog
	def getExportTitle(self):
		return 'Export HL2-SMD Static Mesh'
	
	# default extension for export dialog
	def getExportExtension(self):
		return 'smd'
	
	# find objects to export
	def checkInitState(self):
		if not self.objMesh:
			alertUser('There is no Mesh selected. Select at last a Mesh and optional an Armature', 'Error')
			return False
		if not self.refAction:
			alertUser('No Reference Action found (action named "reference" or "reference.ignore").', 'Error')
			return False
		if self.multiFoldMesh:
			alertUser('Can not export Multi-Fold meshes (Edges used by more than 2 faces).', 'Error')
			return False
		if self.degeneratedFaceCount > 0:
			alertUser('Degenerated Faces found, clean mesh first!', 'Error')
			return False
		return True
	
	# write mesh informations
	def writeMesh(self):
		self.file.write( 'version 1\n' ) # version one file... don't ask me what else could go here
		return True
	
	# write bones
	def writeBones(self):
		if debugLevel > 0: print 'saving bones...'
		self.file.write( 'nodes\n' )
		self.file.write( '0 "%s" -1\n' % ( self.objMesh.data.name ) )
		if self.armature:
			for bone in self.bones:
				armBone = self.armature.bones[ bone.index ]
				bone.restMat = convertMatrix( armBone.getRestMatrix() )
				if armBone.hasParent():
					parentMat = convertMatrix( armBone.getParent().getRestMatrix() )
					parentMat.invert()
					bone.restMat = bone.restMat * parentMat
					parentBoneName = armBone.getParent().name
					for findBone in self.bones:
						if parentBoneName == findBone.name:
							bone.parent = findBone.index
							break
				bone.pos = vector_by_matrix( bone.restMat, Vector([0,0,0]) )
				bone.rot = matrixToEuler( bone.restMat )
				# write bone informations
				self.file.write( '%i "%s" %i\n' % ( bone.index + 1, bone.name, bone.parent + 1 ) )
		self.file.write( 'end\n' )
		self.file.write( 'skeleton\n' )
		self.file.write( 'time 0\n' )
		self.file.write( '0 0.000000 0.000000 0.000000 0.000000 0.000000 0.000000\n' )
		for bone in self.bones:
			wpos = vector_by_matrix( transformDEToHL2, bone.pos )
			wrot = vector_by_matrix( transformDEToHL2, bone.rot )
			if debugLevel > 1: print '- bone', bone.name, 'parent', bone.parent, 'pos', wpos, 'rot', wrot
			self.file.write( '%i %f %f %f %f %f %f\n' % ( bone.index + 1, wpos.x, wpos.y, wpos.z,
				wrot.x * ONE_PI, wrot.y * ONE_PI, wrot.z * ONE_PI) )
		self.file.write( 'end\n' )
		return True
	
	# write faces
	def writeFaces(self):
		if debugLevel > 0: print 'saving faces...'
		self.file.write( 'triangles\n' )
		for face in self.faces:
			# texture
			material = self.mesh.materials[ self.mesh.faces[ face.index ].materialIndex ]
			# write triangle (for both cases)
			if len( face.vertices ) == 3:
				if debugLevel > 2: print '- tri', material.getName(), ':', face.v[2].index, ',', face.v[1].index, ',', face.v[0].index
				self.file.write( '%s\n' % ( material.getName() ) )
				if not self.writeIndexedFace( face, [0,1,2] ):
					return False
			elif len( face.vertices ) == 4:
				if debugLevel > 2: print '- quad', material.getName(), ':', face.v[3].index, ',', face.v[2].index, ',', face.v[1].index, ',', face.v[0].index
				self.file.write( '%s\n' % ( material.getName() ) )
				if not self.writeIndexedFace( face, [0,1,2] ):
					return False
				self.file.write( '%s\n' % ( material.getName() ) )
				if not self.writeIndexedFace( face, [0,2,3] ):
					return False
		self.file.write( 'end\n' )
		return True
	
	# helper function to write one single face using the given indices
	def writeIndexedFace(self, face, indexList):
		for t in indexList:
			meshFace = self.mesh.faces[ face.index ]
			faceVert = self.vertices[ face.vertices[t] ]
			# position
			vec = Vector( meshFace.v[t].co )
			vertPos = vector_by_matrix( transformBlenderToHL2, vec )
			# normal
			vec = Vector( meshFace.v[t].no )
			vertNormal = vector_by_matrix( transformBlenderToHL2, vec )
			# texture coordinates
			vertU = meshFace.uv[t][0]
			vertV = meshFace.uv[t][1]
			# write to file
			self.file.write( '0 %f %f %f %f %f %f %f %f %i' % ( vertPos.x, vertPos.y, vertPos.z,
				vertNormal.x, vertNormal.y, vertNormal.z, vertU, vertV, len( faceVert.weights ) ) )
			for weight in faceVert.weights:
				self.file.write( ' %i %f' % ( weight.index + 1, weight.weight ) )
			self.file.write( '\n' )
		return True
	
	# protected export to file function
	def safeExport(self):
		# preparation work
		self.switchActionTime(self.refAction, 1)
		# write file
		Blender.Window.DrawProgressBar(0.0, 'Writing Mesh Informations')
		if not self.writeMesh():
			return False
		Blender.Window.DrawProgressBar(0.1, 'Writing Bone Informations')
		if not self.writeBones():
			return False
		Blender.Window.DrawProgressBar(0.5, 'Writing Face Informations')
		if not self.writeFaces():
			return False
		Blender.Window.DrawProgressBar(1.0, 'Finished')
		# finished
		return True	



# animation exporter
######################
class HLExporterAnimation ( DEBaseExporter ):
	
	# constructor
	def __init__(self):
		DEBaseExporter.__init__(self)
	
	# title of export dialog
	def getExportTitle(self):
		return 'Export HL2-SMD Animation Sequence'
	
	# default extension for export dialog
	def getExportExtension(self):
		return 'smd'
	
	# find objects to export
	def checkInitState(self):
		if not self.objMesh:
			alertUser('There is no Mesh selected. Select at last a Mesh and optional an Armature', 'Error')
			return False
		if not self.refAction:
			alertUser('No Reference Action found (action named "reference" or "reference.ignore").', 'Error')
			return False
		if self.multiFoldMesh:
			alertUser('Can not export Multi-Fold meshes (Edges used by more than 2 faces).', 'Error')
			return False
		if self.degeneratedFaceCount > 0:
			alertUser('Degenerated Faces found, clean mesh first!', 'Error')
			return False
		return True
	
	# protected export to file function
	def safeExport(self):
		alertUser('Not supported yet!', 'Error')
		return False
		# preparation work
		#self.switchActionTime(self.refAction, 1)
		# finished, we did nothing <.=.<
		#return True	



# create a filename that is matching our file format
def newFName(ext):
  return Blender.Get('filename')[: -len(Blender.Get('filename').split('.', -1)[-1]) ] + ext


# output debugging informations
def outputDebug(message):
	Blender.Draw.PupMenu('Half-Life 2 Export Debug%t|' + message)


# alert user of something
def alertUser(message, title):
	Blender.Draw.PupMenu('Half-Life 2 Export ' + title + '%t|' + message)



# helper function to convert matrix from blender to dragengine
def convertMatrix(matrix):
	# first determine the axes
	axisX = axis_by_matrix( transformBlenderToDE, axis_by_matrix( matrix, Vector([1,0,0]) ) )
	axisY = axis_by_matrix( transformBlenderToDE, axis_by_matrix( matrix, Vector([0,0,1]) ) )
	axisZ = axis_by_matrix( transformBlenderToDE, axis_by_matrix( matrix, Vector([0,1,0]) ) )
	pos = vector_by_matrix( transformBlenderToDE, vector_by_matrix( matrix, Vector([0,0,0]) ) )
	# create epsylon ready matrix
	return Matrix( [axisX.x, axisX.y, axisX.z, 0], [axisY.x, axisY.y, axisY.z, 0], [axisZ.x, axisZ.y, axisZ.z, 0], [pos.x, pos.y, pos.z, 1] )


# helper function calculating the euler angles
# from a given converted matrix
def matrixToEuler(matrix):
	euler = Vector([0, 0, 0])
	if matrix[2][1] > 1:
		euler.x = 90 * ONE_PI
	elif matrix[2][1] < -1:
		euler.x = -90 * ONE_PI
	else:
		euler.x = math.asin(matrix[2][1])
	if euler.x < HALF_PI:
		if euler.x > -HALF_PI:
			euler.z = math.atan2( -matrix[0][1], matrix[1][1] )
			euler.y = math.atan2( -matrix[2][0], matrix[2][2] )
		else:
			# undefined
			euler.z = -math.atan2( matrix[0][2], matrix[0][0] )
			euler.y = 0
	else:
		# undefined
		euler.z = math.atan2( matrix[0][2], matrix[0][0] )
		euler.y = 0
	return Vector([ euler.x / ONE_PI, euler.y / ONE_PI, euler.z / ONE_PI ])

# helper function to transform a vector by a matrix
def vector_by_matrix(m, p):
  return Vector( [ p[0] * m[0][0] + p[1] * m[1][0] + p[2] * m[2][0] + m[3][0],
          p[0] * m[0][1] + p[1] * m[1][1] + p[2] * m[2][1] + m[3][1],
          p[0] * m[0][2] + p[1] * m[1][2] + p[2] * m[2][2] + m[3][2] ] )

# helper function to transform an axis vector by a matrix
def axis_by_matrix(m, p):
  return Vector( [ p[0] * m[0][0] + p[1] * m[1][0] + p[2] * m[2][0],
          p[0] * m[0][1] + p[1] * m[1][1] + p[2] * m[2][1],
          p[0] * m[0][2] + p[1] * m[1][2] + p[2] * m[2][2] ] )


# callback for the file dialog
def fileDialogCallback(filename):
	extension = '.' + exporter.getExportExtension()
	if filename.find(extension, -6) <= 0:
		filename += extension
	exporter.export(filename)

# create a filename that is matching our file format
def newFName():
	return Blender.Get('filename')[: -len(Blender.Get('filename').split('.', -1)[-1]) ] + exporter.getExportExtension()


# debugging informations
# 0 = no debug infos
# 1 = a bit of debug infos
# 2 = even more debug infos
# 3 = heavy debug infos, makes your console explode ;=)
debugLevel = 1

# entry point.
args = __script__['arg']
if args == 'export_static':
	exporter = HLExporterStaticMesh()
	Blender.Window.FileSelector(fileDialogCallback, exporter.getExportTitle(), newFName())
elif args == 'export_anim':
	exporter = HLExporterAnimation()
	Blender.Window.FileSelector(fileDialogCallback, exporter.getExportTitle(), newFName())
else:
	alertUser('Now I am confuzzled. What shall I do with \'%s\'?!' % (args), 'Error')
#Blender.Redraw() # i have no clue why this has no effect at all >.=.<
